@@ -18,7 +18,6 @@ from six.moves.urllib.parse import urlparse
 from flask import Flask, Response, jsonify, request
 from localstack import config
 from localstack.constants import TEST_AWS_ACCOUNT_ID
-from localstack.services import generic_proxy
 from localstack.utils.aws import aws_stack, aws_responses
 from localstack.services.awslambda import lambda_executors
 from localstack.services.awslambda.lambda_executors import (
@@ -38,13 +37,17 @@ from localstack.services.awslambda.lambda_executors import (
     LAMBDA_RUNTIME_RUBY,
     LAMBDA_RUNTIME_RUBY25,
     LAMBDA_RUNTIME_PROVIDED)
-from localstack.utils.common import (to_str, load_file, save_file, TMP_FILES, ensure_readable,
-    mkdir, unzip, is_zip_file, zip_contains_jar_entries, run, short_uid,
-    timestamp_millis, parse_chunked_data, now_utc, safe_requests, FuncThread,
-    isoformat_milliseconds)
+from localstack.services.awslambda.multivalue_transformer import multi_value_dict_for_list
+from localstack.utils.common import (
+    to_str, to_bytes, load_file, save_file, TMP_FILES, ensure_readable, short_uid, json_safe,
+    mkdir, unzip, is_zip_file, zip_contains_jar_entries, run, first_char_to_lower,
+    timestamp_millis, now_utc, safe_requests, FuncThread, isoformat_milliseconds, synchronized)
 from localstack.utils.analytics import event_publisher
+from localstack.utils.http_utils import parse_chunked_data
 from localstack.utils.aws.aws_models import LambdaFunction
-from localstack.utils.cloudwatch.cloudwatch_util import cloudwatched
+
+# logger
+LOG = logging.getLogger(__name__)
 
 APP_NAME = 'lambda_api'
 PATH_ROOT = '/2015-03-31'
@@ -69,6 +72,8 @@ LAMBDA_JAR_FILE_NAME = 'original_lambda_archive.jar'
 
 INVALID_PARAMETER_VALUE_EXCEPTION = 'InvalidParameterValueException'
 
+VERSION_LATEST = '$LATEST'
+
 BATCH_SIZE_RANGES = {
     'kinesis': (100, 10000),
     'dynamodb': (100, 1000),
@@ -78,16 +83,13 @@ BATCH_SIZE_RANGES = {
 app = Flask(APP_NAME)
 
 # map ARN strings to lambda function objects
-arn_to_lambda = {}
+ARN_TO_LAMBDA = {}
 
 # list of event source mappings for the API
-event_source_mappings = []
-
-# logger
-LOG = logging.getLogger(__name__)
+EVENT_SOURCE_MAPPINGS = []
 
 # mutex for access to CWD and ENV
-exec_mutex = threading.Semaphore(1)
+EXEC_MUTEX = threading.RLock(1)
 
 # whether to use Docker for execution
 DO_USE_DOCKER = None
@@ -117,7 +119,7 @@ LAMBDA_EXECUTOR = lambda_executors.AVAILABLE_EXECUTORS.get(config.LAMBDA_EXECUTO
 
 # IAM policy constants
 IAM_POLICY_VERSION = '2012-10-17'
-POLICY_NAME_PATTERN = 'lambda_policy_%s'
+POLICY_NAME_PATTERN = 'lambda_policy_%s_%s'
 
 # Marker name to indicate that a bucket represents the local file system. This is used for testing
 # Serverless applications where we mount the Lambda code directly into the container from the host OS.
@@ -137,10 +139,11 @@ class ClientError(Exception):
 
 
 class LambdaContext(object):
-    def __init__(self, func_details, qualifier=None):
+    def __init__(self, func_details, qualifier=None, client_context=None):
         self.function_name = func_details.name()
         self.function_version = func_details.get_qualifier_version(qualifier)
-
+        if client_context:
+            self.client_context = client_context
         self.invoked_function_arn = func_details.arn()
         if qualifier:
             self.invoked_function_arn += ':' + qualifier
@@ -151,9 +154,9 @@ class LambdaContext(object):
 
 
 def cleanup():
-    global event_source_mappings, arn_to_lambda
-    arn_to_lambda = {}
-    event_source_mappings = []
+    global EVENT_SOURCE_MAPPINGS, ARN_TO_LAMBDA
+    ARN_TO_LAMBDA = {}
+    EVENT_SOURCE_MAPPINGS = []
     LAMBDA_EXECUTOR.cleanup()
 
 
@@ -180,8 +183,8 @@ def check_batch_size_range(source_arn, batch_size=None):
 
 def add_function_mapping(lambda_name, lambda_handler, lambda_cwd=None):
     arn = func_arn(lambda_name)
-    arn_to_lambda[arn].versions.get('$LATEST')['Function'] = lambda_handler
-    arn_to_lambda[arn].cwd = lambda_cwd
+    ARN_TO_LAMBDA[arn].versions.get(VERSION_LATEST)['Function'] = lambda_handler
+    ARN_TO_LAMBDA[arn].cwd = lambda_cwd
 
 
 def add_event_source(function_name, source_arn, enabled, batch_size=None):
@@ -192,18 +195,18 @@ def add_event_source(function_name, source_arn, enabled, batch_size=None):
         'StateTransitionReason': 'User action',
         'LastModified': float(time.mktime(datetime.utcnow().timetuple())),
         'BatchSize': batch_size,
-        'State': 'Enabled' if enabled is True or enabled is None else 'Disabled',
+        'State': 'Enabled' if enabled in [True, None] else 'Disabled',
         'FunctionArn': func_arn(function_name),
         'EventSourceArn': source_arn,
         'LastProcessingResult': 'OK',
         'StartingPosition': LAMBDA_DEFAULT_STARTING_POSITION
     }
-    event_source_mappings.append(mapping)
+    EVENT_SOURCE_MAPPINGS.append(mapping)
     return mapping
 
 
 def update_event_source(uuid_value, function_name, enabled, batch_size):
-    for m in event_source_mappings:
+    for m in EVENT_SOURCE_MAPPINGS:
         if uuid_value == m['UUID']:
             if function_name:
                 m['FunctionArn'] = func_arn(function_name)
@@ -220,12 +223,13 @@ def update_event_source(uuid_value, function_name, enabled, batch_size):
 
 
 def delete_event_source(uuid_value):
-    for i, m in enumerate(event_source_mappings):
+    for i, m in enumerate(EVENT_SOURCE_MAPPINGS):
         if uuid_value == m['UUID']:
-            return event_source_mappings.pop(i)
+            return EVENT_SOURCE_MAPPINGS.pop(i)
     return {}
 
 
+@synchronized(lock=EXEC_MUTEX)
 def use_docker():
     global DO_USE_DOCKER
     if DO_USE_DOCKER is None:
@@ -239,45 +243,83 @@ def use_docker():
     return DO_USE_DOCKER
 
 
-def process_apigateway_invocation(func_arn, path, payload, headers={},
-        resource_path=None, method=None, path_params={},
-        query_string_params={}, request_context={}):
+def get_stage_variables(api_id, stage):
+    api_gateway_client = aws_stack.connect_to_service('apigateway')
+    response = api_gateway_client.get_stage(restApiId=api_id, stageName=stage)
+    return response.get('variables', None)
+
+
+def fix_proxy_path_params(path_params):
+    proxy_path_param_value = path_params.get('proxy+')
+    if not proxy_path_param_value:
+        return
+    del path_params['proxy+']
+    path_params['proxy'] = proxy_path_param_value
+
+
+def message_attributes_to_lower(message_attrs):
+    """ Convert message attribute details (first characters) to lower case (e.g., stringValue, dataType). """
+    message_attrs = message_attrs or {}
+    for _, attr in message_attrs.items():
+        if not isinstance(attr, dict):
+            continue
+        for key, value in dict(attr).items():
+            attr[first_char_to_lower(key)] = attr.pop(key)
+    return message_attrs
+
+
+def process_apigateway_invocation(func_arn, path, payload, stage, api_id, headers={},
+                                  resource_path=None, method=None, path_params={},
+                                  query_string_params=None, request_context={}):
     try:
         resource_path = resource_path or path
+        path_params = dict(path_params)
+        fix_proxy_path_params(path_params)
         event = {
             'path': path,
             'headers': dict(headers),
-            'pathParameters': dict(path_params),
+            'multiValueHeaders': multi_value_dict_for_list(headers),
+            'pathParameters': path_params,
             'body': payload,
             'isBase64Encoded': False,
             'resource': resource_path,
             'httpMethod': method,
             'queryStringParameters': query_string_params,
+            'multiValueQueryStringParameters': multi_value_dict_for_list(query_string_params),
             'requestContext': request_context,
-            'stageVariables': {}  # TODO
+            'stageVariables': get_stage_variables(api_id, stage),
         }
-        return run_lambda(event=event, context={}, func_arn=func_arn)
+        return run_lambda(event=event, context={}, func_arn=func_arn,
+            asynchronous=not config.SYNCHRONOUS_API_GATEWAY_EVENTS)
     except Exception as e:
         LOG.warning('Unable to run Lambda function on API Gateway message: %s %s' % (e, traceback.format_exc()))
 
 
-def process_sns_notification(func_arn, topic_arn, subscriptionArn, message, message_attributes, subject='',):
+def process_sns_notification(func_arn, topic_arn, subscription_arn, message,
+        message_attributes, unsubscribe_url, subject='',):
     event = {
         'Records': [{
             'EventSource': 'localstack:sns',
             'EventVersion': '1.0',
-            'EventSubscriptionArn': subscriptionArn,
+            'EventSubscriptionArn': subscription_arn,
             'Sns': {
                 'Type': 'Notification',
+                'MessageId': str(uuid.uuid4()),
                 'TopicArn': topic_arn,
                 'Subject': subject,
                 'Message': message,
                 'Timestamp': timestamp_millis(),
+                'SignatureVersion': '1',
+                # TODO Add a more sophisticated solution with an actual signature
+                # Hardcoded
+                'Signature': 'EXAMPLEpH+..',
+                'SigningCertUrl': 'https://sns.us-east-1.amazonaws.com/SimpleNotificationService-000000000.pem',
+                'UnsubscribeUrl': unsubscribe_url,
                 'MessageAttributes': message_attributes
             }
         }]
     }
-    return run_lambda(event=event, context={}, func_arn=func_arn, asynchronous=True)
+    return run_lambda(event=event, context={}, func_arn=func_arn, asynchronous=not config.SYNCHRONOUS_SNS_EVENTS)
 
 
 def process_kinesis_records(records, stream_name):
@@ -303,9 +345,7 @@ def process_kinesis_records(records, stream_name):
                         for rec in chunk
                     ]
                 }
-
-                run_lambda(event=event, context={}, func_arn=arn)
-
+                run_lambda(event=event, context={}, func_arn=arn, asynchronous=not config.SYNCHRONOUS_KINESIS_EVENTS)
     except Exception as e:
         LOG.warning('Unable to run Lambda function on Kinesis records: %s %s' % (e, traceback.format_exc()))
 
@@ -328,6 +368,7 @@ def start_lambda_sqs_listener():
 
         records = []
         for msg in messages:
+            message_attrs = message_attributes_to_lower(msg.get('MessageAttributes'))
             records.append({
                 'body': msg['Body'],
                 'receiptHandle': msg['ReceiptHandle'],
@@ -337,7 +378,7 @@ def start_lambda_sqs_listener():
                 'awsRegion': region,
                 'messageId': msg['MessageId'],
                 'attributes': msg.get('Attributes', {}),
-                'messageAttributes': msg.get('MessageAttributes', {}),
+                'messageAttributes': message_attrs,
                 'md5OfMessageAttributes': msg.get('MD5OfMessageAttributes'),
                 'sqs': True,
             })
@@ -398,11 +439,11 @@ def process_sqs_message(queue_name, region_name=None):
         queue_arn = aws_stack.sqs_queue_arn(queue_name, region_name=region_name)
         sources = get_event_sources(source_arn=queue_arn)
         arns = [s.get('FunctionArn') for s in sources]
-        LOG.debug('Found %s source mappings for event from SQS queue %s: %s' % (len(arns), queue_arn, arns))
         source = (sources or [None])[0]
         if not source:
             return False
 
+        LOG.debug('Found %s source mappings for event from SQS queue %s: %s' % (len(arns), queue_arn, arns))
         start_lambda_sqs_listener()
         return True
     except Exception as e:
@@ -411,7 +452,7 @@ def process_sqs_message(queue_name, region_name=None):
 
 def get_event_sources(func_name=None, source_arn=None):
     result = []
-    for m in event_source_mappings:
+    for m in EVENT_SOURCE_MAPPINGS:
         if not func_name or (m['FunctionArn'] in [func_name, func_arn(func_name)]):
             if _arn_match(mapped=m['EventSourceArn'], searched=source_arn):
                 result.append(m)
@@ -436,18 +477,18 @@ def _arn_match(mapped, searched):
 
 
 def get_function_version(arn, version):
-    func = arn_to_lambda.get(arn)
+    func = ARN_TO_LAMBDA.get(arn)
     return format_func_details(func, version=version, always_add_version=True)
 
 
 def publish_new_function_version(arn):
-    func_details = arn_to_lambda.get(arn)
+    func_details = ARN_TO_LAMBDA.get(arn)
     versions = func_details.versions
     last_version = func_details.max_version()
     versions[str(last_version + 1)] = {
-        'CodeSize': versions.get('$LATEST').get('CodeSize'),
-        'CodeSha256': versions.get('$LATEST').get('CodeSha256'),
-        'Function': versions.get('$LATEST').get('Function'),
+        'CodeSize': versions.get(VERSION_LATEST).get('CodeSize'),
+        'CodeSha256': versions.get(VERSION_LATEST).get('CodeSha256'),
+        'Function': versions.get(VERSION_LATEST).get('Function'),
         'RevisionId': str(uuid.uuid4())
     }
     return get_function_version(arn, str(last_version + 1))
@@ -455,7 +496,7 @@ def publish_new_function_version(arn):
 
 def do_list_versions(arn):
     return sorted([get_function_version(arn, version) for version in
-                   arn_to_lambda.get(arn).versions.keys()], key=lambda k: str(k.get('Version')))
+                   ARN_TO_LAMBDA.get(arn).versions.keys()], key=lambda k: str(k.get('Version')))
 
 
 def do_update_alias(arn, alias, version, description=None):
@@ -466,11 +507,10 @@ def do_update_alias(arn, alias, version, description=None):
         'Description': description or '',
         'RevisionId': str(uuid.uuid4())
     }
-    arn_to_lambda.get(arn).aliases[alias] = new_alias
+    ARN_TO_LAMBDA.get(arn).aliases[alias] = new_alias
     return new_alias
 
 
-@cloudwatched('lambda')
 def run_lambda(event, context, func_arn, version=None, suppress_output=False, asynchronous=False, callback=None):
     if suppress_output:
         stdout_ = sys.stdout
@@ -480,15 +520,24 @@ def run_lambda(event, context, func_arn, version=None, suppress_output=False, as
         sys.stderr = stream
     try:
         func_arn = aws_stack.fix_arn(func_arn)
-        func_details = arn_to_lambda.get(func_arn)
+        func_details = ARN_TO_LAMBDA.get(func_arn)
         if not func_details:
             return not_found_error(msg='The resource specified in the request does not exist.')
-        if not context:
-            context = LambdaContext(func_details, version)
+
+        context = LambdaContext(func_details, version, context)
+
         result = LAMBDA_EXECUTOR.execute(func_arn, func_details, event, context=context,
-            version=version, asynchronous=asynchronous, callback=callback)
+                                         version=version, asynchronous=asynchronous, callback=callback)
+
     except Exception as e:
-        return error_response('Error executing Lambda function %s: %s %s' % (func_arn, e, traceback.format_exc()))
+        exc_type, exc_value, exc_traceback = sys.exc_info()
+        response = {
+            'errorType': str(exc_type.__name__),
+            'errorMessage': str(e),
+            'stackTrace': traceback.format_tb(exc_traceback)
+        }
+        LOG.info('Error executing Lambda function %s: %s %s' % (func_arn, e, traceback.format_exc()))
+        return Response(json.dumps(response), status=500)
     finally:
         if suppress_output:
             sys.stdout = stdout_
@@ -498,7 +547,7 @@ def run_lambda(event, context, func_arn, version=None, suppress_output=False, as
 
 def exec_lambda_code(script, handler_function='handler', lambda_cwd=None, lambda_env=None):
     if lambda_cwd or lambda_env:
-        exec_mutex.acquire()
+        EXEC_MUTEX.acquire()
         if lambda_cwd:
             previous_cwd = os.getcwd()
             os.chdir(lambda_cwd)
@@ -537,7 +586,7 @@ def exec_lambda_code(script, handler_function='handler', lambda_cwd=None, lambda
                 sys.path.pop(0)
             if lambda_env:
                 os.environ = previous_env
-            exec_mutex.release()
+            EXEC_MUTEX.release()
     return module_vars[handler_function]
 
 
@@ -622,7 +671,7 @@ def get_java_handler(zip_file_content, main_file, func_details=None):
 def set_archive_code(code, lambda_name, zip_file_content=None):
     # get metadata
     lambda_arn = func_arn(lambda_name)
-    lambda_details = arn_to_lambda[lambda_arn]
+    lambda_details = ARN_TO_LAMBDA[lambda_arn]
     is_local_mount = code.get('S3Bucket') == BUCKET_MARKER_LOCAL
 
     if is_local_mount and config.LAMBDA_REMOTE_DOCKER:
@@ -644,8 +693,9 @@ def set_archive_code(code, lambda_name, zip_file_content=None):
 
     # Save the zip file to a temporary file that the lambda executors can reference
     code_sha_256 = base64.standard_b64encode(hashlib.sha256(zip_file_content).digest())
-    lambda_details.get_version('$LATEST')['CodeSize'] = len(zip_file_content)
-    lambda_details.get_version('$LATEST')['CodeSha256'] = code_sha_256.decode('utf-8')
+    latest_version = lambda_details.get_version(VERSION_LATEST)
+    latest_version['CodeSize'] = len(zip_file_content)
+    latest_version['CodeSha256'] = code_sha_256.decode('utf-8')
     tmp_dir = '%s/zipfile.%s' % (config.TMP_FOLDER, short_uid())
     mkdir(tmp_dir)
     tmp_file = '%s/%s' % (tmp_dir, LAMBDA_ZIP_FILE_NAME)
@@ -661,7 +711,7 @@ def set_function_code(code, lambda_name, lambda_cwd=None):
             'Node.js, Golang, and .Net Core Lambdas currently require LAMBDA_EXECUTOR=docker') % lambda_name)
 
     arn = func_arn(lambda_name)
-    lambda_details = arn_to_lambda[arn]
+    lambda_details = ARN_TO_LAMBDA[arn]
     runtime = lambda_details.runtime
     lambda_environment = lambda_details.envvars
     handler_name = lambda_details.handler = lambda_details.handler or LAMBDA_DEFAULT_HANDLER
@@ -742,26 +792,39 @@ def set_function_code(code, lambda_name, lambda_cwd=None):
 
 def do_list_functions():
     funcs = []
-    for f_arn, func in arn_to_lambda.items():
+    this_region = aws_stack.get_region()
+    for f_arn, func in ARN_TO_LAMBDA.items():
         if type(func) != LambdaFunction:
             continue
+
+        # filter out functions of current region
+        func_region = f_arn.split(':')[3]
+        if func_region != this_region:
+            continue
+
         func_name = f_arn.split(':function:')[-1]
         arn = func_arn(func_name)
-        func_details = arn_to_lambda.get(arn)
+        func_details = ARN_TO_LAMBDA.get(arn)
         if not func_details:
             # this can happen if we're accessing Lambdas from a different region (ARN mismatch)
             continue
-        funcs.append(format_func_details(func_details))
+
+        details = format_func_details(func_details)
+        details['Tags'] = func.tags
+
+        funcs.append(details)
     return funcs
 
 
 def format_func_details(func_details, version=None, always_add_version=False):
-    version = version or '$LATEST'
+    version = version or VERSION_LATEST
     func_version = func_details.get_version(version)
     result = {
         'CodeSha256': func_version.get('CodeSha256'),
         'Role': func_details.role,
+        'KMSKeyArn': func_details.kms_key_arn,
         'Version': version,
+        'VpcConfig': func_details.vpc_config,
         'FunctionArn': func_details.arn(),
         'FunctionName': func_details.name(),
         'CodeSize': func_version.get('CodeSize'),
@@ -783,7 +846,7 @@ def format_func_details(func_details, version=None, always_add_version=False):
         result['Environment'] = {
             'Variables': func_details.envvars
         }
-    if (always_add_version or version != '$LATEST') and len(result['FunctionArn'].split(':')) <= 7:
+    if (always_add_version or version != VERSION_LATEST) and len(result['FunctionArn'].split(':')) <= 7:
         result['FunctionArn'] += ':%s' % version
     return result
 
@@ -873,11 +936,12 @@ def create_function():
         event_publisher.fire_event(event_publisher.EVENT_LAMBDA_CREATE_FUNC,
             payload={'n': event_publisher.get_hash(lambda_name)})
         arn = func_arn(lambda_name)
-        if arn in arn_to_lambda:
+        if arn in ARN_TO_LAMBDA:
             return error_response('Function already exist: %s' %
                 lambda_name, 409, error_type='ResourceConflictException')
-        arn_to_lambda[arn] = func_details = LambdaFunction(arn)
-        func_details.versions = {'$LATEST': {'RevisionId': str(uuid.uuid4())}}
+        ARN_TO_LAMBDA[arn] = func_details = LambdaFunction(arn)
+        func_details.versions = {VERSION_LATEST: {'RevisionId': str(uuid.uuid4())}}
+        func_details.vpc_config = data.get('VpcConfig', {})
         func_details.last_modified = isoformat_milliseconds(datetime.utcnow()) + '+0000'
         func_details.description = data.get('Description', '')
         func_details.handler = data['Handler']
@@ -886,22 +950,23 @@ def create_function():
         func_details.tags = data.get('Tags', {})
         func_details.timeout = data.get('Timeout', LAMBDA_DEFAULT_TIMEOUT)
         func_details.role = data['Role']
+        func_details.kms_key_arn = data.get('KMSKeyArn')
         func_details.memory_size = data.get('MemorySize')
         func_details.code = data['Code']
         func_details.set_dead_letter_config(data)
         result = set_function_code(func_details.code, lambda_name)
         if isinstance(result, Response):
-            del arn_to_lambda[arn]
+            del ARN_TO_LAMBDA[arn]
             return result
         # remove content from code attribute, if present
         func_details.code.pop('ZipFile', None)
         # prepare result
         result.update(format_func_details(func_details))
-        if data.get('Publish', False):
+        if data.get('Publish'):
             result['Version'] = publish_new_function_version(arn)['Version']
         return jsonify(result or {})
     except Exception as e:
-        arn_to_lambda.pop(arn, None)
+        ARN_TO_LAMBDA.pop(arn, None)
         if isinstance(e, ClientError):
             return e.get_response()
         return error_response('Unknown error: %s %s' % (e, traceback.format_exc()))
@@ -925,9 +990,10 @@ def get_function(function):
                 'Configuration': func,
                 'Code': {
                     'Location': '%s/code' % request.url
-                }
+                },
+                'Tags': func['Tags']
             }
-            lambda_details = arn_to_lambda.get(func['FunctionArn'])
+            lambda_details = ARN_TO_LAMBDA.get(func['FunctionArn'])
             if lambda_details.concurrency is not None:
                 result['Concurrency'] = lambda_details.concurrency
             return jsonify(result)
@@ -944,8 +1010,9 @@ def list_functions():
               in: body
     """
     funcs = do_list_functions()
-    result = {}
-    result['Functions'] = funcs
+    result = {
+        'Functions': funcs
+    }
     return jsonify(result)
 
 
@@ -964,17 +1031,17 @@ def delete_function(function):
     LAMBDA_EXECUTOR.cleanup(arn)
 
     try:
-        arn_to_lambda.pop(arn)
+        ARN_TO_LAMBDA.pop(arn)
     except KeyError:
         return not_found_error(func_arn(function))
 
     event_publisher.fire_event(event_publisher.EVENT_LAMBDA_DELETE_FUNC,
         payload={'n': event_publisher.get_hash(function)})
     i = 0
-    while i < len(event_source_mappings):
-        mapping = event_source_mappings[i]
+    while i < len(EVENT_SOURCE_MAPPINGS):
+        mapping = EVENT_SOURCE_MAPPINGS[i]
         if mapping['FunctionArn'] == arn:
-            del event_source_mappings[i]
+            del EVENT_SOURCE_MAPPINGS[i]
             i -= 1
         i += 1
     result = {}
@@ -993,7 +1060,7 @@ def update_function_code(function):
     data = json.loads(to_str(request.data))
     result = set_function_code(data, function)
     arn = func_arn(function)
-    func_details = arn_to_lambda.get(arn)
+    func_details = ARN_TO_LAMBDA.get(arn)
     result.update(format_func_details(func_details))
     if isinstance(result, Response):
         return result
@@ -1008,7 +1075,7 @@ def get_function_code(function):
         parameters:
     """
     arn = func_arn(function)
-    lambda_cwd = arn_to_lambda[arn].cwd
+    lambda_cwd = ARN_TO_LAMBDA[arn].cwd
     tmp_file = '%s/%s' % (lambda_cwd, LAMBDA_ZIP_FILE_NAME)
     return Response(load_file(tmp_file, mode='rb'),
             mimetype='application/zip',
@@ -1023,7 +1090,7 @@ def get_function_configuration(function):
         parameters:
     """
     arn = func_arn(function)
-    lambda_details = arn_to_lambda.get(arn)
+    lambda_details = ARN_TO_LAMBDA.get(arn)
     if not lambda_details:
         return not_found_error(arn)
     result = format_func_details(lambda_details)
@@ -1045,7 +1112,7 @@ def update_function_configuration(function):
     # Stop/remove any containers that this arn uses.
     LAMBDA_EXECUTOR.cleanup(arn)
 
-    lambda_details = arn_to_lambda.get(arn)
+    lambda_details = ARN_TO_LAMBDA.get(arn)
     if not lambda_details:
         return error_response('Unable to find Lambda function ARN "%s"' % arn,
             404, error_type='ResourceNotFoundException')
@@ -1080,8 +1147,9 @@ def add_permission(function):
             'Resource': func_arn(function)
         }]
     }
-    iam_client.create_policy(PolicyName=POLICY_NAME_PATTERN % function,
-        PolicyDocument=json.dumps(policy), Description='Policy for Lambda function "%s"' % function)
+    iam_client.create_policy(PolicyName=POLICY_NAME_PATTERN % (function, sid),
+                             PolicyDocument=json.dumps(policy),
+                             Description='Policy for Lambda function "%s"' % function)
     result = {'Statement': sid}
     return jsonify(result)
 
@@ -1131,7 +1199,6 @@ def invoke_function(function):
         arn = m.group(1)
     else:
         qualifier = request.args.get('Qualifier')
-
     data = request.get_data()
     if data:
         data = to_str(data)
@@ -1149,15 +1216,17 @@ def invoke_function(function):
     invocation_type = request.environ.get('HTTP_X_AMZ_INVOCATION_TYPE', 'RequestResponse')
 
     def _create_response(result, status_code=200, headers={}):
-        """ Create the final response for the given invocation result """
-        if isinstance(result, Response):
-            return result
+        """ Create the final response for the given invocation result. """
         details = {
             'StatusCode': status_code,
             'Payload': result,
             'Headers': headers
         }
-        if isinstance(result, dict):
+        if isinstance(result, Response):
+            details['Payload'] = to_str(result.data)
+            if result.status_code >= 400:
+                details['FunctionError'] = 'Unhandled'
+        elif isinstance(result, dict):
             for key in ('StatusCode', 'Payload', 'FunctionError'):
                 if result.get(key):
                     details[key] = result[key]
@@ -1173,9 +1242,12 @@ def invoke_function(function):
         # Set error headers
         if details.get('FunctionError'):
             details['Headers']['X-Amz-Function-Error'] = str(details['FunctionError'])
+        details['Headers']['X-Amz-Log-Result'] = base64.b64encode(to_bytes(''))  # TODO add logs!
+        details['Headers']['X-Amz-Executed-Version'] = str(qualifier or VERSION_LATEST)
         # Construct response object
         response_obj = details['Payload']
         if was_json or isinstance(response_obj, JSON_START_TYPES):
+            response_obj = json_safe(response_obj)
             response_obj = jsonify(response_obj)
             details['Headers']['Content-Type'] = 'application/json'
         else:
@@ -1185,9 +1257,9 @@ def invoke_function(function):
 
     # check if this lambda function exists
     not_found = None
-    if arn not in arn_to_lambda:
+    if arn not in ARN_TO_LAMBDA:
         not_found = not_found_error(arn)
-    elif qualifier and not arn_to_lambda.get(arn).qualifier_exists(qualifier):
+    elif qualifier and not ARN_TO_LAMBDA.get(arn).qualifier_exists(qualifier):
         not_found = not_found_error('{0}:{1}'.format(arn, qualifier))
 
     if not_found:
@@ -1197,7 +1269,8 @@ def invoke_function(function):
         return not_found
 
     if invocation_type == 'RequestResponse':
-        result = run_lambda(asynchronous=False, func_arn=arn, event=data, context={}, version=qualifier)
+        result = run_lambda(asynchronous=False, func_arn=arn, event=data,
+                            context=request.headers.get('X-Amz-Client-Context'), version=qualifier)
         return _create_response(result)
     elif invocation_type == 'Event':
         run_lambda(asynchronous=True, func_arn=arn, event=data, context={}, version=qualifier)
@@ -1210,7 +1283,7 @@ def invoke_function(function):
 
 
 @app.route('%s/event-source-mappings/' % PATH_ROOT, methods=['GET'])
-def list_event_source_mappings():
+def list_EVENT_SOURCE_MAPPINGS():
     """ List event source mappings
         ---
         operationId: 'listEventSourceMappings'
@@ -1218,7 +1291,7 @@ def list_event_source_mappings():
     event_source_arn = request.args.get('EventSourceArn')
     function_name = request.args.get('FunctionName')
 
-    mappings = event_source_mappings
+    mappings = EVENT_SOURCE_MAPPINGS
     if event_source_arn:
         mappings = [m for m in mappings if event_source_arn == m.get('EventSourceArn')]
     if function_name:
@@ -1240,7 +1313,7 @@ def get_event_source_mapping(mapping_uuid):
             - name: 'request'
               in: body
     """
-    mappings = event_source_mappings
+    mappings = EVENT_SOURCE_MAPPINGS
     mappings = [m for m in mappings if mapping_uuid == m.get('UUID')]
 
     if len(mappings) == 0:
@@ -1309,7 +1382,7 @@ def delete_event_source_mapping(mapping_uuid):
 @app.route('%s/functions/<function>/versions' % PATH_ROOT, methods=['POST'])
 def publish_version(function):
     arn = func_arn(function)
-    if arn not in arn_to_lambda:
+    if arn not in ARN_TO_LAMBDA:
         return not_found_error(arn)
     return jsonify(publish_new_function_version(arn))
 
@@ -1317,7 +1390,7 @@ def publish_version(function):
 @app.route('%s/functions/<function>/versions' % PATH_ROOT, methods=['GET'])
 def list_versions(function):
     arn = func_arn(function)
-    if arn not in arn_to_lambda:
+    if arn not in ARN_TO_LAMBDA:
         return not_found_error(arn)
     return jsonify({'Versions': do_list_versions(arn)})
 
@@ -1325,11 +1398,11 @@ def list_versions(function):
 @app.route('%s/functions/<function>/aliases' % PATH_ROOT, methods=['POST'])
 def create_alias(function):
     arn = func_arn(function)
-    if arn not in arn_to_lambda:
+    if arn not in ARN_TO_LAMBDA:
         return not_found_error(arn)
     data = json.loads(request.data)
     alias = data.get('Name')
-    if alias in arn_to_lambda.get(arn).aliases:
+    if alias in ARN_TO_LAMBDA.get(arn).aliases:
         return error_response('Alias already exists: %s' % arn + ':' + alias, 404,
                               error_type='ResourceConflictException')
     version = data.get('FunctionVersion')
@@ -1340,11 +1413,11 @@ def create_alias(function):
 @app.route('%s/functions/<function>/aliases/<name>' % PATH_ROOT, methods=['PUT'])
 def update_alias(function, name):
     arn = func_arn(function)
-    if arn not in arn_to_lambda:
+    if arn not in ARN_TO_LAMBDA:
         return not_found_error(arn)
-    if name not in arn_to_lambda.get(arn).aliases:
+    if name not in ARN_TO_LAMBDA.get(arn).aliases:
         return not_found_error(msg='Alias not found: %s:%s' % (arn, name))
-    current_alias = arn_to_lambda.get(arn).aliases.get(name)
+    current_alias = ARN_TO_LAMBDA.get(arn).aliases.get(name)
     data = json.loads(request.data)
     version = data.get('FunctionVersion') or current_alias.get('FunctionVersion')
     description = data.get('Description') or current_alias.get('Description')
@@ -1354,19 +1427,19 @@ def update_alias(function, name):
 @app.route('%s/functions/<function>/aliases/<name>' % PATH_ROOT, methods=['GET'])
 def get_alias(function, name):
     arn = func_arn(function)
-    if arn not in arn_to_lambda:
+    if arn not in ARN_TO_LAMBDA:
         return not_found_error(arn)
-    if name not in arn_to_lambda.get(arn).aliases:
+    if name not in ARN_TO_LAMBDA.get(arn).aliases:
         return not_found_error(msg='Alias not found: %s:%s' % (arn, name))
-    return jsonify(arn_to_lambda.get(arn).aliases.get(name))
+    return jsonify(ARN_TO_LAMBDA.get(arn).aliases.get(name))
 
 
 @app.route('%s/functions/<function>/aliases' % PATH_ROOT, methods=['GET'])
 def list_aliases(function):
     arn = func_arn(function)
-    if arn not in arn_to_lambda:
+    if arn not in ARN_TO_LAMBDA:
         return not_found_error(arn)
-    return jsonify({'Aliases': sorted(arn_to_lambda.get(arn).aliases.values(),
+    return jsonify({'Aliases': sorted(ARN_TO_LAMBDA.get(arn).aliases.values(),
                                       key=lambda x: x['Name'])})
 
 
@@ -1377,7 +1450,7 @@ def put_concurrency(version, function):
     # https://docs.aws.amazon.com/lambda/latest/dg/API_PutFunctionConcurrency.html
     arn = func_arn(function)
     data = json.loads(request.data)
-    lambda_details = arn_to_lambda.get(arn)
+    lambda_details = ARN_TO_LAMBDA.get(arn)
     if not lambda_details:
         return not_found_error(arn)
     lambda_details.concurrency = data
@@ -1386,7 +1459,7 @@ def put_concurrency(version, function):
 
 @app.route('/<version>/tags/<arn>', methods=['GET'])
 def list_tags(version, arn):
-    func_details = arn_to_lambda.get(arn)
+    func_details = ARN_TO_LAMBDA.get(arn)
     if not func_details:
         return not_found_error(arn)
     result = {'Tags': func_details.tags}
@@ -1398,7 +1471,7 @@ def tag_resource(version, arn):
     data = json.loads(request.data)
     tags = data.get('Tags', {})
     if tags:
-        func_details = arn_to_lambda.get(arn)
+        func_details = ARN_TO_LAMBDA.get(arn)
         if not func_details:
             return not_found_error(arn)
         if func_details:
@@ -1409,7 +1482,7 @@ def tag_resource(version, arn):
 @app.route('/<version>/tags/<arn>', methods=['DELETE'])
 def untag_resource(version, arn):
     tag_keys = request.args.getlist('tagKeys')
-    func_details = arn_to_lambda.get(arn)
+    func_details = ARN_TO_LAMBDA.get(arn)
     if not func_details:
         return not_found_error(arn)
     for tag_key in tag_keys:
@@ -1432,7 +1505,7 @@ def put_function_event_invoke_config(function):
     """
     data = json.loads(to_str(request.data))
     function_arn = func_arn(function)
-    lambda_obj = arn_to_lambda[function_arn]
+    lambda_obj = ARN_TO_LAMBDA[function_arn]
     response = lambda_obj.put_function_event_invoke_config(data)
 
     return jsonify({
@@ -1466,7 +1539,7 @@ def get_function_event_invoke_config(function):
     """
     try:
         function_arn = func_arn(function)
-        lambda_obj = arn_to_lambda[function_arn]
+        lambda_obj = ARN_TO_LAMBDA[function_arn]
     except Exception as e:
         return error_response(str(e), 400)
 
@@ -1475,6 +1548,8 @@ def get_function_event_invoke_config(function):
 
 
 def serve(port, quiet=True):
+    from localstack.services import generic_proxy  # moved here to fix circular import errors
+
     # initialize the Lambda executor
     LAMBDA_EXECUTOR.startup()
 
